import logging
from collections import defaultdict
from typing import DefaultDict, Dict, Iterable, List, Optional, Type

from rxn.utilities.strings import remove_postfix

from .actions import Action, InvalidAction
from .conversion_utils import (
    ActionToStringError,
    SingleActionConverter,
    StringToActionError,
    UnsupportedType,
)
from .converter_interface import ActionStringConverter
from .default_converters import default_action_converters

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class ReadableConverter(ActionStringConverter):
    """
    Converter for synthesis actions to and from readable text.

    This class relies on registering converters. For the string-to-action conversion,
    there may be several converters for the same name. In this case, the priority
    follows which was registered first.
    """

    def __init__(
        self,
        separator: str = "; ",
        end_mark: str = ".",
        single_action_converters: Optional[Iterable[SingleActionConverter]] = None,  # type: ignore[type-arg]
    ):
        """
        Args:
            separator: string that will be inserted between the action strings.
            end_mark: string that will be added at the end of the action string.
            single_action_converters: converters to register. Defaults to the
                defaults ones from the default_converters.py module.
        """
        super().__init__()

        self.separator = separator
        self.end_mark = end_mark

        # If a compound name / duration contains the separator, the string
        # will be modified by adding a no-break space between the first and
        # second character.
        self.separator_substitute = self.separator[:1] + "\u200C" + self.separator[1:]

        self.mapping_action: Dict[Type[Action], SingleActionConverter] = {}  # type: ignore[type-arg]
        # Note: defaultdict, as there may be several action classes with the same name.
        self.mapping_uppercase_name: DefaultDict[str, List[SingleActionConverter]] = defaultdict(list)  # type: ignore[type-arg]

        if single_action_converters is None:
            single_action_converters = default_action_converters()

        self.register_many(single_action_converters)

    def register_many(self, action_converters: Iterable[SingleActionConverter]) -> None:  # type: ignore[type-arg]
        """Register several single-action converters."""
        for action_converter in action_converters:
            self.register(action_converter)

    def register(self, action_converter: SingleActionConverter) -> None:  # type: ignore[type-arg]
        """Register a single-action converter."""
        self.mapping_action[action_converter.action_type] = action_converter
        self.mapping_uppercase_name[action_converter.first_word()].append(
            action_converter
        )

    def action_type_supported(self, action_type: str) -> bool:
        all_action_names = (
            action_class.__name__ for action_class in self.mapping_action.keys()
        )
        return any(action_type == name for name in all_action_names)

    def actions_to_string(self, actions: List[Action]) -> str:
        action_strings = (self.action_to_string(a) for a in actions)
        return self.separator.join(action_strings) + self.end_mark

    def string_to_actions(
        self,
        action_string: str,
        wrap_errors_into_invalidaction: bool = False,
    ) -> List[Action]:
        # remove last dot (or other end mark)
        if self.end_mark:
            try:
                action_string = remove_postfix(
                    action_string, self.end_mark, raise_if_missing=True
                )
            except ValueError as e:
                logger.warning(str(e))

        if not action_string:
            return []

        action_strings = action_string.split(self.separator)

        actions: List[Action] = []
        for action_string in action_strings:
            try:
                action = self.string_to_action(action_string)
            except StringToActionError as e:
                if not wrap_errors_into_invalidaction:
                    raise
                action = InvalidAction(str(e))
            actions.append(action)
        return actions

    def action_to_string(self, action: Action) -> str:
        try:
            action_converter = self.mapping_action[action.__class__]
        except KeyError as e:
            raise UnsupportedType(action.__class__) from e

        try:
            action_string = action_converter.action_to_text(action)
        except Exception as e:
            raise ActionToStringError(
                f"Conversion to string failed for {action}"
            ) from e

        # replace normal space by no-break space if there is '; ' in the string,
        # as it would lead to an error in the back-conversion
        action_string = action_string.replace(self.separator, self.separator_substitute)

        return action_string

    def string_to_action(self, action_string: str) -> Action:
        """
        Convert a string for a single action into an action.

        Raises:
            StringToActionError: in case of error.
        """
        # replace no-break space (probably introduced in action_to_string) by normal space
        action_string = action_string.replace(self.separator_substitute, self.separator)

        action_type = action_string.split(" ", 1)[0]
        single_action_converters = self.mapping_uppercase_name[action_type]
        if len(single_action_converters) == 0:
            # i.e. could not find converter for action
            raise StringToActionError(
                action_string, f'Can not find converter for "{action_string}".'
            )

        exceptions = []
        for converter in single_action_converters:
            try:
                action: Action = converter.text_to_action(action_string)
                return action
            except Exception as e:
                exceptions.append(e)

        # If hasn't returned by now: all the converters failed
        raise StringToActionError(
            action_string,
            f"Conversion failed for {action_string}: {'; '.join(str(e) for e in exceptions)}",
        )



from pathlib import Path
from typing import Tuple

import click
from rxn.utilities.files import load_list_from_file

from paragraph2actions.analysis import (
    action_string_validity,
    full_sentence_accuracy,
    levenshtein_similarity,
    modified_bleu,
    partial_accuracy,
)


@click.command()
@click.option(
    "--ground_truth_file",
    "-g",
    required=True,
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    help="File containing the ground truth",
)
@click.option(
    "--prediction_files",
    "-p",
    multiple=True,
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    help="File containing the translations to compare with the ground truth",
)
def main(ground_truth_file: Path, prediction_files: Tuple[Path, ...]) -> None:
    """Calculate metrics for predictions generated by one or several translation models"""

    ground_truth = load_list_from_file(ground_truth_file)

    predictions = []
    for prediction_file in prediction_files:
        predictions.append(load_list_from_file(prediction_file))

    for filename, p in zip(prediction_files, predictions):
        print(filename)
        print("Full sentence accuracy, pr:", full_sentence_accuracy(ground_truth, p))
        print("String validity, pr:", action_string_validity(p))
        print("BLEU, pr:", modified_bleu(ground_truth, p))
        print("Levenshtein, pr:", levenshtein_similarity(ground_truth, p))
        print("100% accuracy, pr:", partial_accuracy(ground_truth, p, 1.0))
        print("90% accuracy, pr:", partial_accuracy(ground_truth, p, 0.9))
        print("75% accuracy, pr:", partial_accuracy(ground_truth, p, 0.75))
        print()


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
from typing import List

from ..actions import Action


class ActionPostprocessor(ABC):
    """
    Interface for postprocessing on a list of actions.

    In this context, postprocessing means any change to the actions (merging, removal, edition, etc.)
    ActionProcessor classes can be applied to both sentences or to full paragraphs.
    """

    @abstractmethod
    def postprocess(self, actions: List[Action]) -> List[Action]:
        """
        Args:
            actions: list of actions to postprocess

        Returns:
            New (postprocessed) list of actions
        """



from abc import ABC
from typing import List, Optional

import attr


@attr.s(auto_attribs=True)
class Chemical:
    """
    Substance with attached quantities.

    Quantities are optional, and several may be given simultaneously (f.i. ['1 mmol', '2.0 g', '2.0 mL']'), therefore
    they are saved as a list.
    """

    name: str
    quantity: List[str] = attr.Factory(list)


@attr.s(auto_attribs=True)
class Action(ABC):
    """
    Base class for an action.

    See https://github.ibm.com/rxn/action_sequences/wiki/Data-format-for-synthesis-actions for
    basic details about available actions.
    """

    @property
    def action_name(self) -> str:
        return self.__class__.__name__


@attr.s(auto_attribs=True)
class InvalidAction(Action):
    error: str = ""


@attr.s(auto_attribs=True)
class Add(Action):
    material: Chemical
    dropwise: bool = False
    temperature: Optional[str] = None
    atmosphere: Optional[str] = None
    duration: Optional[str] = None


@attr.s(auto_attribs=True)
class CollectLayer(Action):
    """
    Attributes:
        layer: which layer to keep ("aqueous" or "organic")
    """

    layer: str

    def __attrs_post_init__(self) -> None:
        if self.layer not in ["aqueous", "organic"]:
            raise ValueError('layer must be equal to "aqueous" or "organic"')


@attr.s(auto_attribs=True)
class Concentrate(Action):
    pass


@attr.s(auto_attribs=True)
class Degas(Action):
    gas: Optional[str]
    duration: Optional[str] = None


@attr.s(auto_attribs=True)
class DrySolid(Action):
    """Dry a solid under air or vacuum.

    For drying under vacuum, the atmosphere variable should contain the string 'vacuum'.
    For drying on air, the atmosphere variable should contain the string 'air'.
    For other atmospheres, the corresponding gas name should be given ('N2', 'argon', etc.).
    """

    duration: Optional[str] = None
    temperature: Optional[str] = None
    atmosphere: Optional[str] = None


@attr.s(auto_attribs=True)
class DrySolution(Action):
    """Dry an organic solution with a desiccant"""

    material: Optional[str]


@attr.s(auto_attribs=True)
class Extract(Action):
    solvent: Chemical
    repetitions: int = 1


@attr.s(auto_attribs=True)
class Filter(Action):
    """
    Filtration action, possibly with information about what phase to keep ('filtrate' or 'precipitate')
    """

    phase_to_keep: Optional[str] = None

    def __attrs_post_init__(self) -> None:
        if self.phase_to_keep is not None and self.phase_to_keep not in [
            "filtrate",
            "precipitate",
        ]:
            raise ValueError(
                'phase_to_keep must be equal to "filtrate" or "precipitate"'
            )


@attr.s(auto_attribs=True)
class FollowOtherProcedure(Action):
    """
    Fake action for sentences that refer to another experimental procedure.
    """


@attr.s(auto_attribs=True)
class MakeSolution(Action):
    """
    Action to make a solution out of a list of compounds.
    This action is usually followed by another action using it (Add, Quench, etc.).
    """

    materials: List[Chemical]

    def __attrs_post_init__(self) -> None:
        if len(self.materials) < 2:
            raise ValueError(
                f"MakeSolution requires at least two components (actual: {len(self.materials)}"
            )


@attr.s(auto_attribs=True)
class Microwave(Action):
    duration: Optional[str] = None
    temperature: Optional[str] = None


@attr.s(auto_attribs=True)
class OtherLanguage(Action):
    """
    Fake action for sentences that are not in English.
    """


@attr.s(auto_attribs=True)
class Partition(Action):
    material_1: Chemical
    material_2: Chemical


@attr.s(auto_attribs=True)
class PH(Action):
    material: Chemical
    ph: Optional[str] = None
    dropwise: bool = False
    temperature: Optional[str] = None


@attr.s(auto_attribs=True)
class PhaseSeparation(Action):
    pass


@attr.s(auto_attribs=True)
class Purify(Action):
    pass


@attr.s(auto_attribs=True)
class Quench(Action):
    material: Chemical
    dropwise: bool = False
    temperature: Optional[str] = None


@attr.s(auto_attribs=True)
class Recrystallize(Action):
    solvent: Chemical


@attr.s(auto_attribs=True)
class Reflux(Action):
    duration: Optional[str] = None
    dean_stark: bool = False
    atmosphere: Optional[str] = None


@attr.s(auto_attribs=True)
class SetTemperature(Action):
    """
    If there is a duration given with cooling/heating, use "Stir" instead
    """

    temperature: str


@attr.s(auto_attribs=True)
class Sonicate(Action):
    duration: Optional[str] = None
    temperature: Optional[str] = None


@attr.s(auto_attribs=True)
class Stir(Action):
    duration: Optional[str] = None
    temperature: Optional[str] = None
    atmosphere: Optional[str] = None


@attr.s(auto_attribs=True)
class Triturate(Action):
    solvent: Chemical


@attr.s(auto_attribs=True)
class Wait(Action):
    """
    NB: "Wait" as an action can be ambiguous depending on the context.
    It seldom means "waiting without doing anything", but is often "continue what was before", at least in Pistachio.
    """

    duration: str
    temperature: Optional[str] = None


@attr.s(auto_attribs=True)
class Wash(Action):
    material: Chemical
    repetitions: int = 1


@attr.s(auto_attribs=True)
class Yield(Action):
    material: Chemical


@attr.s(auto_attribs=True)
class NoAction(Action):
    """
    Fake action for sentences that actually have no action.
    """



import inspect
import itertools
from typing import Any, Callable, Iterable, Iterator, List, Optional, Sequence, Tuple

import attr

from .actions import Action, Chemical
from .actions import __dict__ as actions_module_dict  # type: ignore

temperature_attribute_names = ["temperature"]
duration_attribute_names = ["duration"]
atmosphere_attribute_names = ["atmosphere"]
compound_attribute_names = [
    "gas",
    "material",
    "material_1",
    "material_2",
    "materials",
    "solvent",
]


def get_all_action_types() -> List[str]:
    """Returns a list of the available action names"""

    actions: List[str] = []
    for (
        name,
        val,
    ) in actions_module_dict.items():  # iterate through every module's attributes
        if inspect.isclass(val) and issubclass(val, Action) and val != Action:
            actions.append(name)

    return actions


def extract_chemicals(
    actions: Iterable[Action], ignore_sln: bool = False
) -> List[Chemical]:
    """
    Returns a list of all the chemicals present in a sequence of actions.
    """
    chemicals = []

    for a in actions:
        for value in a.__dict__.values():
            if isinstance(value, Chemical):
                chemicals.append(value)
            if isinstance(value, list):
                for v in value:
                    if isinstance(v, Chemical):
                        chemicals.append(v)

    if ignore_sln:
        chemicals = [chemical for chemical in chemicals if chemical.name != "SLN"]

    return chemicals


def actions_with_compounds(actions: Iterable[Action]) -> List[Tuple[Action, str]]:
    """
    In a list of actions, looks for the ones having a compound (that may be
    set or not).

    Returns them as a list of actions paired with the name of the member
    variable corresponding to the compound. Includes both compounds specified
    as strings and as instances of Chemical.
    """

    return _actions_with_attribute_names(actions, compound_attribute_names)


def extract_compound_names(
    actions: Iterable[Action], ignore_sln: bool = True
) -> List[str]:
    """
    Returns a list of all the compounds (as strings) present in a sequence of
    actions.

    Will not only return the ones corresponding to the Chemical instances
    extracted by extract_chemicals, but also the ones that are present only
    as strings such as Degas.gas or DrySolution.material.
    """

    compounds = []

    for action, attribute_name in actions_with_compounds(actions):
        value = getattr(action, attribute_name)
        if value is None:
            continue
        elif isinstance(value, str):
            compounds.append(value)
        elif isinstance(value, Chemical):
            compounds.append(value.name)
        elif isinstance(value, list):
            for element in value:
                if isinstance(element, Chemical):
                    compounds.append(element.name)

    if ignore_sln:
        compounds = [compound for compound in compounds if compound != "SLN"]

    return compounds


def _actions_with_attribute_names(
    actions: Iterable[Action], attribute_names: Sequence[str]
) -> List[Tuple[Action, str]]:
    """
    In a list of actions, looks for the ones having a given attribute names
    (that may be set or not).

    Useful for actions_with_temperatures or actions_with_durations, for
    instance. Returns them as a list of actions paired with the name of the
    member variable.
    """
    tuples: List[Tuple[Action, str]] = []

    for a in actions:
        for attribute_name in attribute_names:
            if hasattr(a, attribute_name):
                tuples.append((a, attribute_name))

    return tuples


def _apply_to_actions_with_attribute_names(
    actions: Iterable[Action],
    attribute_names: Sequence[str],
    fn: Callable[[str], Optional[str]],
) -> None:
    """
    Applies a function of all the duration fields present in the given actions,
    to update their values.

    Args:
        actions: actions on which to apply fn
        attribute_names: attribute names to apply to
        fn: function to apply to the durations, basically doing
            `action.<attribute_name> = fn(action.<attribute_name>)`
    """
    actions_and_attributes = _actions_with_attribute_names(actions, attribute_names)

    for a, attribute_name in actions_and_attributes:
        value = getattr(a, attribute_name, None)

        # do nothing if there is no value (f.i. optional variables)
        if value is None:
            continue

        # set the new value
        new_value = fn(value)
        setattr(a, attribute_name, new_value)


def apply_to_temperatures(actions: Iterable[Action], fn: Callable[[str], str]) -> None:
    """
    Applies a function of all the temperature fields present in the given
    actions, to update their values.

    Args:
        actions: actions on which to apply fn
        fn: function to apply to the temperatures, basically doing
            `action.temperature = fn(action.temperature)`
    """
    _apply_to_actions_with_attribute_names(actions, temperature_attribute_names, fn)


def apply_to_durations(actions: Iterable[Action], fn: Callable[[str], str]) -> None:
    """
    Applies a function of all the duration fields present in the given actions,
    to update their values.

    Args:
        actions: actions on which to apply fn
        fn: function to apply to the durations, basically doing
            `action.duration = fn(action.duration)`
    """
    _apply_to_actions_with_attribute_names(actions, duration_attribute_names, fn)


def apply_to_atmospheres(
    actions: Iterable[Action], fn: Callable[[str], Optional[str]]
) -> None:
    """
    Applies a function of all the atmosphere fields present in the given actions,
    to update their values.

    Args:
        actions: actions on which to apply fn
        fn: function to apply to the atmospheres, basically doing
            `action.atmosphere = fn(action.atmosphere)`
    """
    _apply_to_actions_with_attribute_names(actions, atmosphere_attribute_names, fn)


def actions_with_temperatures(actions: Iterable[Action]) -> List[Tuple[Action, str]]:
    """
    In a list of actions, looks for the ones having a temperature (that may be
    set or not).

    Returns them as a list of actions paired with the name of the member
    variable corresponding to the temperature.
    """

    return _actions_with_attribute_names(actions, temperature_attribute_names)


def actions_with_durations(actions: Iterable[Action]) -> List[Tuple[Action, str]]:
    """
    In a list of actions, looks for the ones having a duration (that may be
    set or not).

    Returns them as a list of actions paired with the name of the member
    variable corresponding to the duration.
    """

    return _actions_with_attribute_names(actions, duration_attribute_names)


def extract_temperatures(actions: Iterable[Action]) -> List[str]:
    """
    Returns a list of all the temperatures present in a sequence of actions.
    """
    temperatures = []

    for action, attribute_name in actions_with_temperatures(actions):
        value = getattr(action, attribute_name)
        if value is not None:
            temperatures.append(value)

    return temperatures


def extract_durations(actions: Iterable[Action]) -> List[str]:
    """
    Returns a list of all the durations present in a sequence of actions.
    """
    durations = []

    for action, attribute_name in actions_with_durations(actions):
        value = getattr(action, attribute_name)
        if value is not None:
            durations.append(value)

    return durations


def remove_quantities(actions: Iterable[Action]) -> None:
    """Remove the quantities of a list of actions in-place."""
    chemicals = extract_chemicals(actions)
    for chemical in chemicals:
        chemical.quantity = []


def pairwise(s: List[Any]) -> Iterator[Tuple[Any, Any]]:
    """
    Iterates over neighbors in a list.

    s -> (s0,s1), (s1,s2), (s2, s3), ...

    From https://stackoverflow.com/a/5434936
    """

    a, b = itertools.tee(s)
    next(b, None)
    return zip(a, b)


def all_identical(sequence: Sequence[Any]) -> bool:
    return all(s == sequence[0] for s in sequence)


@attr.s(auto_attribs=True)
class Sentence:
    """
    Sentence from a synthesis recipe with the corresponding actions.
    """

    text: str
    actions: List[Action]



import copy
import logging
from typing import List

from ..actions import Action, Add, MakeSolution
from .action_postprocessor import ActionPostprocessor

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class InitialMakesolutionPostprocessor(ActionPostprocessor):
    """
    Replaces MakeSolution actions at the beginning of a sequence by Add actions.
    """

    def postprocess(self, actions: List[Action]) -> List[Action]:
        actions = copy.deepcopy(actions)

        if not self.has_makesolution_at_beginning(actions):
            return actions

        makesolution = actions[0]
        add = actions[1]
        assert isinstance(makesolution, MakeSolution)
        assert isinstance(add, Add)

        atmosphere = add.atmosphere
        temperature = add.temperature
        if add.dropwise:
            logger.warning("Dropwise addition of initial MakeSolution is ignored")
        if add.duration is not None:
            logger.warning("Duration of addition of initial MakeSolution is ignored")

        adds = [
            Add(m, temperature=temperature, atmosphere=atmosphere)
            for m in makesolution.materials
        ]

        # replace "MakeSolution X; Add SLN" by the new Adds
        actions[0:2] = adds

        return actions

    def has_makesolution_at_beginning(self, actions: List[Action]) -> bool:
        if len(actions) < 2:
            return False

        if not isinstance(actions[0], MakeSolution):
            return False

        add = actions[1]
        if not isinstance(add, Add):
            return False

        if add.material.name != "SLN":
            return False

        return True



from abc import ABC, abstractmethod

from paragraph2actions.misc import TextWithActions


class Augmenter(ABC):
    """
    Interface for data augmentation for the action translation dataset.
    """

    @abstractmethod
    def augment(self, sample: TextWithActions) -> TextWithActions:
        """
        Generates a new sample that has potentially been data-augmented.
        """



from typing import List

from ..actions import Action, DrySolution, Filter
from .action_postprocessor import ActionPostprocessor


class DrysolutionPostprocessor(ActionPostprocessor):
    """
    Ensures that a DrySolution action is followed by a Filter action.

    Often in paragraphs the filtering step is not mentioned. The data was
    analyzed to check if sometimes it comes just a few steps later. Most of
    these cases, however, relate to another filtering step. For instance, it is
    common to dry the solution with desiccant, then concentrate, and then filter
    again, this time to keep the precipitate.
    The safest is therefore probably to always ensure that there is a Filter
    action after DrySolution.
    """

    def postprocess(self, actions: List[Action]) -> List[Action]:
        postprocessed: List[Action] = []

        for i, action in enumerate(actions):
            postprocessed.append(action)

            # If the current action is not a DrySolution: go to the next one
            if not isinstance(action, DrySolution):
                continue

            # If the next action is not Filter: add it.
            next_action = actions[i + 1] if i < len(actions) - 1 else None
            if not isinstance(next_action, Filter):
                postprocessed.append(Filter("filtrate"))

        return postprocessed



import copy
from typing import List, Type

from ..actions import Action, Concentrate, DrySolid, DrySolution, Filter
from ..utils import pairwise
from .action_postprocessor import ActionPostprocessor


class FilterPostprocessor(ActionPostprocessor):
    """
    Sets what to keep (filtrate vs precipitate) in a Filter action based on the following action.
    """

    def __init__(self) -> None:
        self.pre_filtrate_classes: List[Type[Action]] = [DrySolution]
        self.post_filtrate_classes: List[Type[Action]] = [Concentrate, DrySolution]
        self.pre_precipitate_classes: List[Type[Action]] = []
        self.post_precipitate_classes: List[Type[Action]] = [DrySolid]

    def postprocess(self, actions: List[Action]) -> List[Action]:
        actions = copy.deepcopy(actions)

        for a, b in pairwise(actions):
            self.update_based_on_following_action(filter_action=a, following_action=b)
            self.update_based_on_previous_action(filter_action=b, previous_action=a)

        return actions

    def update_based_on_previous_action(
        self, filter_action: Action, previous_action: Action
    ) -> None:
        self.update_based_on_other_action(
            filter_action,
            previous_action,
            self.pre_filtrate_classes,
            self.pre_precipitate_classes,
        )

    def update_based_on_following_action(
        self, filter_action: Action, following_action: Action
    ) -> None:
        self.update_based_on_other_action(
            filter_action,
            following_action,
            self.post_filtrate_classes,
            self.post_precipitate_classes,
        )

    def update_based_on_other_action(
        self,
        filter_action: Action,
        other_action: Action,
        filtrate_related_classes: List[Type[Action]],
        precipitate_related_classes: List[Type[Action]],
    ) -> None:
        if (
            not isinstance(filter_action, Filter)
            or filter_action.phase_to_keep is not None
        ):
            return

        if any(isinstance(other_action, cls) for cls in filtrate_related_classes):
            filter_action.phase_to_keep = "filtrate"

        if any(isinstance(other_action, cls) for cls in precipitate_related_classes):
            filter_action.phase_to_keep = "precipitate"



from typing import Dict, List, Tuple, Type

import rxn.utilities.attrs

from .actions import Action


def get_variables(cls: Type[Action]) -> List[str]:
    """
    For a given action class, returns the names of the variables.
    """
    return rxn.utilities.attrs.get_variables(cls)


def get_variables_and_types(cls: Type[Action]) -> List[Tuple[str, Type]]:  # type: ignore[type-arg]
    """
    For a given action class, returns the names of the variables and their types.
    """
    return rxn.utilities.attrs.get_variables_and_types(cls)


def get_variables_and_type_names(cls: Type[Action]) -> List[Tuple[str, str]]:
    """
    For a given action class, returns the names of the variables and their type names.
    """
    return rxn.utilities.attrs.get_variables_and_type_names(cls)


def get_variable_type(cls: Type[Action], variable_name: str) -> Type:  # type: ignore[type-arg]
    """
    For a given action class and variable name, returns the type of that variable.
    """
    for name, variable_type in get_variables_and_types(cls):
        if name == variable_name:
            return variable_type
    raise ValueError(f'No variable "{variable_name}" for class {cls.__name__}')


def inspect_all_variables() -> Dict[str, List[str]]:
    """
    Returns a dictionary of all the action classes names with the corresponding variables.
    """
    return {
        action_cls.__name__: get_variables(action_cls)
        for action_cls in Action.__subclasses__()
    }


def inspect_all_variables_and_types() -> Dict[str, List[Tuple[str, str]]]:
    """
    Returns a dictionary of all the action classes names with the corresponding variables, and types.
    """
    return {
        action_cls.__name__: get_variables_and_type_names(action_cls)
        for action_cls in Action.__subclasses__()
    }



import sentencepiece as spm


class SentencePieceTokenizer:
    """
    Wrapper for the SentencePiece tokenizer
    """

    def __init__(self, model_file: str):
        self.sp = spm.SentencePieceProcessor()
        self.sp.Load(model_file)

    def tokenize(self, sentence: str) -> str:
        tokens = self.sp.EncodeAsPieces(sentence)
        tokenized = " ".join(tokens)
        return tokenized

    def detokenize(self, sentence: str) -> str:
        tokens = sentence.split(" ")
        detokenized: str = self.sp.DecodePieces(tokens)
        return detokenized



import copy
from typing import List

from paragraph2actions.misc import TextWithActions

from ..utils import extract_chemicals
from .substitution_augmenter import SubstitutionAugmenter


class CompoundQuantityAugmenter(SubstitutionAugmenter):
    """
    Augments data by substituting compound quantities.

    Adding or removing quantities is not considered, since this is more complex to change in the text.
    """

    def __init__(self, probability: float, quantities: List[str]):
        """
        Args:
            probability: probability with which to switch the compound name
            quantities: list of quantities to use for substitution
        """
        super().__init__(probability=probability, values=quantities)

    def augment(self, sample: TextWithActions) -> TextWithActions:
        sample = copy.deepcopy(sample)

        chemicals = extract_chemicals(sample.actions)

        quantity_blocks = [c.quantity for c in chemicals]
        all_quantities = [
            q for quantity_block in quantity_blocks for q in quantity_block
        ]

        unique_quantities = set(all_quantities)

        # remove quantities that are comprised in others; with this, if both '1.0 g' and
        # '21.0 g' are present as quantities, we will never substitute the short one.
        for q in list(unique_quantities):
            if any(q in quantity for quantity in unique_quantities if q != quantity):
                unique_quantities.remove(q)

        # For each quantity, try substitution
        for q in unique_quantities:
            if not self.random_draw_passes() or q not in sample.text:
                continue

            new_quantity = self.draw_value()
            sample.text = sample.text.replace(q, new_quantity)

            for quantity_block in quantity_blocks:
                for i in range(len(quantity_block)):
                    if q == quantity_block[i]:
                        quantity_block[i] = new_quantity

        return sample



import logging
import random
from itertools import cycle, islice
from pathlib import Path
from typing import Iterable, List

from rxn.utilities.containers import remove_duplicates
from rxn.utilities.files import load_list_from_file

from .augmentation.compound_name_augmenter import CompoundNameAugmenter
from .augmentation.compound_quantity_augmenter import CompoundQuantityAugmenter
from .augmentation.duration_augmenter import DurationAugmenter
from .augmentation.temperature_augmenter import TemperatureAugmenter
from .conversion_utils import SingleActionConverter
from .misc import TextWithActions, load_samples, save_samples
from .readable_converter import ReadableConverter

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def augment_annotations(
    data_dir: Path,
    value_lists_dir: Path,
    output_dir: Path,
    single_action_converters: Iterable[SingleActionConverter],  # type: ignore[type-arg]
    n_augmentations: int,
) -> None:
    readable_converter = ReadableConverter(
        single_action_converters=single_action_converters
    )

    train_items = load_samples(
        text_file=str(data_dir / "src-train.txt"),
        actions_file=str(data_dir / "tgt-train.txt"),
        converter=readable_converter,
    )

    logger.info("Samples in train set:", len(train_items))

    # data augmentation for train split
    compound_names = load_list_from_file(str(value_lists_dir / "compound_names.txt"))
    quantities = load_list_from_file(str(value_lists_dir / "quantities.txt"))
    durations = load_list_from_file(str(value_lists_dir / "durations.txt"))
    temperatures = load_list_from_file(str(value_lists_dir / "temperatures.txt"))

    cna = CompoundNameAugmenter(0.5, compound_names)
    cqa = CompoundQuantityAugmenter(0.5, quantities)
    da = DurationAugmenter(0.5, durations)
    ta = TemperatureAugmenter(0.5, temperatures)

    def augment(sample: TextWithActions) -> TextWithActions:
        sample = cna.augment(sample)
        sample = cqa.augment(sample)
        sample = da.augment(sample)
        sample = ta.augment(sample)
        return sample

    augmented_samples = [
        augment(sample)
        for sample in islice(cycle(train_items), n_augmentations * len(train_items))
    ]

    random.shuffle(augmented_samples)
    logger.info("Samples in augmented train set:", len(augmented_samples))

    augmented_samples_unique = remove_duplicates(
        augmented_samples, key=lambda x: x.text
    )
    logger.info(
        "Samples in augmented train set after removing duplicates:",
        len(augmented_samples_unique),
    )

    def save_to_file(sample_list: List[TextWithActions], subset_name: str) -> None:
        src_file = f"{output_dir}/src-{subset_name}.txt"
        tgt_file = f"{output_dir}/tgt-{subset_name}.txt"

        logger.info(f"Saving {len(sample_list)} samples to {src_file} and {tgt_file}")
        save_samples(
            samples=sample_list,
            converter=readable_converter,
            text_file=src_file,
            actions_file=tgt_file,
        )

    output_dir.mkdir(exist_ok=True)
    save_to_file(augmented_samples_unique, "train-augmented-unique")
    save_to_file(augmented_samples, "train-augmented")



import json
import logging
from typing import Iterable, List

import attr
from rxn.utilities.files import PathLike

from .actions import Action
from .converter_interface import ActionStringConverter

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


@attr.s(auto_attribs=True)
class TextWithActions:
    """
    Sentence (or paragraph) and its corresponding actions.
    """

    text: str
    actions: List[Action]


def load_samples(
    text_file: PathLike, actions_file: PathLike, converter: ActionStringConverter
) -> List[TextWithActions]:
    """
    Loads samples of sentences with corresponding actions from files.
    Sentences and actions are loaded from different files, which corresponds to the format of OpenNMT.

    Args:
        text_file: where to load the original text from
        actions_file: where to load the actions from
        converter: how to convert from the string representation to the actions
    """

    with open(text_file, "rt") as f_sents, open(actions_file, "rt") as f_actns:
        sentences = [line.strip() for line in f_sents]
        actions_lists = [converter.string_to_actions(line.strip()) for line in f_actns]

    assert len(sentences) == len(actions_lists)
    return [
        TextWithActions(sentence, actions)
        for sentence, actions in zip(sentences, actions_lists)
    ]


def save_samples(
    samples: Iterable[TextWithActions],
    converter: ActionStringConverter,
    text_file: PathLike,
    actions_file: PathLike,
) -> None:
    """
    Saves samples of sentences with corresponding actions to files.
    Sentences and actions are saved to different files, which corresponds to the input required by OpenNMT.

    Args:
        samples: samples to save to the file
        converter: how to convert the actions to a string representation
        text_file: where to save the original text
        actions_file: where to save the actions
    """

    with open(text_file, "wt") as f_src, open(actions_file, "wt") as f_tgt:
        for s in samples:
            f_src.write(f"{s.text}\n")
            f_tgt.write(f"{converter.actions_to_string(s.actions)}\n")


def load_samples_from_json(
    annotated_file: PathLike, converter: ActionStringConverter
) -> List[TextWithActions]:
    """
    Loads samples of sentences with corresponding actions from the JSON file.

    Args:
        annotated_file: where to load the text and actions from.
        converter: how to convert from the string representation to the actions.
    """

    samples = []
    with open(annotated_file, "rt") as f:
        for line in f:
            d = json.loads(line)
            approved = d.get("approved", True)
            if not approved:
                continue

            try:
                actions = converter.string_to_actions(d["actions"])
            except Exception as e:
                logger.error(f"Error converting action \"{d['actions']}\": {e}")
                continue

            samples.append(
                TextWithActions(
                    text=d["sentence"],
                    actions=actions,
                )
            )

    return samples



import copy
from typing import List

from ..actions import Action, NoAction
from .action_postprocessor import ActionPostprocessor


class NoActionPostprocessor(ActionPostprocessor):
    """
    Postprocessor that removes "NoAction" from a list of actions.
    """

    def postprocess(self, actions: List[Action]) -> List[Action]:
        actions = copy.deepcopy(actions)
        return [a for a in actions if not isinstance(a, NoAction)]



import copy
from typing import Callable, List, Optional

from ..actions import Action
from ..utils import apply_to_temperatures, extract_temperatures
from .action_postprocessor import ActionPostprocessor


class SameTemperaturePostprocessor(ActionPostprocessor):
    """
    Postprocessor that converts "same temperature" to an actual temperature by
    looking backwards in the text.
    """

    def __init__(self) -> None:
        self.same_temperature_names = {"same temperature"}

    def postprocess(self, actions: List[Action]) -> List[Action]:
        actions = copy.deepcopy(actions)

        # Check if it is necessary at all to change something.
        # If not, return early
        temperatures = set(extract_temperatures(actions))
        if not temperatures.intersection(self.same_temperature_names):
            return actions

        for i, action in enumerate(actions):
            # Get the potential substitue in case "same temperature" is in the current action.
            substitute = self.get_last_temperature(actions[:i])
            if substitute is not None:
                # Create the callback for temperature replacement
                fn = self.create_replace_function(substitute)
                apply_to_temperatures([action], fn)

        return actions

    def get_last_temperature(self, actions: List[Action]) -> Optional[str]:
        """Get the last temperature in a list of actions"""
        temperatures = extract_temperatures(actions)
        valid_temperatures = [
            t for t in temperatures if t not in self.same_temperature_names
        ]
        if not valid_temperatures:
            return None
        return valid_temperatures[-1]

    def create_replace_function(self, substitute: str) -> Callable[[str], str]:
        """
        Creates a callback that can be used to replace a temperature member variable that specifies "same temperature".
        """

        def replace_fn(temperature: str) -> str:
            if temperature in self.same_temperature_names:
                return substitute
            return temperature

        return replace_fn



import re
from typing import List

from .sentence_splitter import SentenceSplitter


class DotSplitter(SentenceSplitter):
    """
    Splits sentences where there are dots followed by spaces.
    """

    def __init__(self) -> None:
        super().__init__(
            split_sentences_at_newlines=True, add_full_stop_if_missing=True
        )

    def _split_impl(self, text: str) -> List[str]:
        # replace multiple spaces / tabs by one single space
        text = re.sub(r"\s+", " ", text)

        # split at ". "
        sentences = text.split(". ")

        # Add the dots back
        sentences = [self._maybe_add_full_stop(s) for s in sentences]

        return sentences

    def _maybe_add_full_stop(self, sentence: str) -> str:
        if sentence.endswith("."):
            return sentence
        return sentence + "."



from typing import List, Optional

from .actions import Action, InvalidAction, NoAction
from .utils import Sentence


class SentencePostprocessor:
    """
    Given sentences with corresponding actions, will post-process them:
    * "NoAction" for empty sentences? which ones?
    * Keep invalid actions?
    * Remove unknown actions?
    """

    def __init__(
        self,
        consider_straightforward_empty_sentences_as_noaction: bool = True,
        remove_sentences_with_invalid_actions: bool = True,
        remove_empty_sentences: bool = True,
    ):
        """
        Args:
            consider_straightforward_empty_sentences_as_noaction: For simple sentences with no action, add "NoAction" and keep them
            remove_sentences_with_invalid_actions: remove sentences where the conversion failed
            remove_empty_sentences: whether to keep empty sentences (NB: NoAction is not considered to be empty)
        """

        self.consider_straightforward_empty_sentences_as_noaction = (
            consider_straightforward_empty_sentences_as_noaction
        )
        self.remove_sentences_with_invalid_actions = (
            remove_sentences_with_invalid_actions
        )
        self.remove_empty_sentences = remove_empty_sentences

        self.no_action_keywords = [
            "MS",
            "found:",
            "Found:",
            "ppm",
            "ESI",
            "NMR",
            "m/z",
            "retention",
            "Retention",
            "HPLC",
            "TLC",
            "M+H",
            "M+1",
            "mp:",
            "Mp:",
        ]
        self.no_action_length = 30

    def process(self, sentence: Sentence) -> Optional[Sentence]:
        """
        Post-processes a sentence

        Returns:
            The post-processed sentence, or None if it should not be included in the dataset
        """

        def replace_actions(actions: List[Action]) -> Sentence:
            return Sentence(text=sentence.text, actions=actions)

        if self.remove_sentences_with_invalid_actions:
            if any(isinstance(action, InvalidAction) for action in sentence.actions):
                return None

        if self.consider_straightforward_empty_sentences_as_noaction:
            if len(sentence.actions) == 0 and self.empty_sentence_is_straightforward(
                sentence.text
            ):
                return replace_actions([NoAction()])

        if self.remove_empty_sentences and len(sentence.actions) == 0:
            return None

        # If we got here, no post-processing is needed and we return the same actions
        return replace_actions(sentence.actions)

    def empty_sentence_is_straightforward(self, sentence_text: str) -> bool:
        """
        Returns True if we are reasonably certain that no actions are included
        """
        # TODO: there's likely an error below: should it be ``in sentence_text.split(" ")``?
        has_analysis_keyword = any(w in sentence_text for w in self.no_action_keywords)
        short_sentence = len(sentence_text) < self.no_action_length

        return has_analysis_keyword or short_sentence



from typing import List

from .action_attribute_augmenter import ActionAttributeAugmenter


class TemperatureAugmenter(ActionAttributeAugmenter):
    """
    Augments data by substituting action temperatures.
    """

    def __init__(self, probability: float, temperatures: List[str]):
        """
        Args:
            probability: probability with which to switch the compound name
            temperatures: list of temperatures to use for substitution
        """
        super().__init__(
            probability=probability, values=temperatures, attribute_name="temperature"
        )



import logging
from typing import Iterable, List, Optional, Union

import attr

from .actions import Action
from .converter_interface import ActionStringConverter
from .readable_converter import ReadableConverter
from .sentence_splitting.dot_splitter import DotSplitter
from .sentence_splitting.sentence_splitter import (
    SentenceSplitter,
    SentenceSplittingError,
)
from .translator import Translator

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


@attr.s(auto_attribs=True)
class Sentence:
    """
    Sentence from a synthesis recipe with the corresponding actions.
    """

    text: str
    actions: List[Action]


@attr.s(auto_attribs=True)
class Paragraph:
    """
    Recipe paragraph and the corresponding actions (available through the contained sentences).
    """

    text: str
    sentences: List[Sentence]

    @property
    def actions(self) -> List[Action]:
        return [action for sentence in self.sentences for action in sentence.actions]

    @property
    def sentence_texts(self) -> List[str]:
        return [s.text for s in self.sentences]


class ParagraphTranslator:
    """
    Translates paragraphs into series of actions.
    """

    def __init__(
        self,
        translation_model: Union[str, Iterable[str]],
        sentencepiece_model: str,
        sentence_splitter: Optional[SentenceSplitter] = None,
        action_string_converter: Optional[ActionStringConverter] = None,
        wrap_errors_into_invalidaction: bool = True,
    ):
        """
        Translates a paragraph and returns the action representation.

        Args:
            translation_model: path(s) to translation model(s)
            sentencepiece_model: path to SentencePiece model
            sentence_splitter: splits the sentences, defaults to ``DotSplitter``.
            action_string_converter: converter to get the actual actions,
                defaults to the ReadableConverter
        """

        self.translator = Translator(
            translation_model=translation_model, sentencepiece_model=sentencepiece_model
        )
        self.converter = (
            action_string_converter
            if action_string_converter is not None
            else ReadableConverter()
        )

        self.sentence_splitter = (
            sentence_splitter if sentence_splitter is not None else DotSplitter()
        )
        # splitter to use when the sentence_splitter fails
        self.fallback_splitter = DotSplitter()

        self.wrap_errors_into_invalidaction = wrap_errors_into_invalidaction

    def extract_paragraph(self, text: str) -> Paragraph:
        """
        Translates a paragraph and returns the action representation.
        """
        sentences = self.split_sentences(text)
        action_strings = self.translator.translate_sentences(sentences)

        actions_per_sentence = [
            self.converter.string_to_actions(
                action_string,
                wrap_errors_into_invalidaction=self.wrap_errors_into_invalidaction,
            )
            for action_string in action_strings
        ]

        paragraph = Paragraph(
            text=text,
            sentences=[
                Sentence(text=s, actions=a)
                for s, a in zip(sentences, actions_per_sentence)
            ],
        )
        return paragraph

    def extract_actions(self, text: str) -> List[Action]:
        """
        Translates a paragraph and returns the action representation.
        """
        p = self.extract_paragraph(text)
        return p.actions

    def split_sentences(self, text: str) -> List[str]:
        try:
            return self.sentence_splitter.split(text)
        except SentenceSplittingError:
            logger.warning(
                "Error during splitting of paragraph in sentences; using fallback splitter."
            )
            return self.fallback_splitter.split(text)



from typing import List

from ..actions import Action
from .action_postprocessor import ActionPostprocessor


class PostprocessorCombiner(ActionPostprocessor):
    """
    Combines multiple other postprocessors
    """

    def __init__(self, postprocessors: List[ActionPostprocessor]):
        self.postprocessors = postprocessors

    def postprocess(self, actions: List[Action]) -> List[Action]:
        for p in self.postprocessors:
            actions = p.postprocess(actions)
        return actions



import random
from abc import ABC
from typing import List

from .augmenter import Augmenter


class SubstitutionAugmenter(Augmenter, ABC):
    """
    Base class for data augmentation relying on substitution.
    """

    def __init__(self, probability: float, values: List[str]):
        """
        Args:
            probability: probability with which to switch a value when prompted
            values: list of values to use for substitution
        """
        self.probability = probability
        self.values = list(values)

        assert 0 <= self.probability <= 1
        assert len(self.values) > 0

    def random_draw_passes(self) -> bool:
        return random.uniform(0, 1) < self.probability

    def draw_value(self) -> str:
        return random.choice(self.values)



import copy
from typing import List

from ..actions import Action, Microwave, Reflux, Stir
from .action_postprocessor import ActionPostprocessor


class DuplicateActionsPostprocessor(ActionPostprocessor):
    """
    Postprocessor that removes repeated actions, if they are not
    type of Stir, Reflux or Microwave.
    """

    def postprocess(self, actions: List[Action]) -> List[Action]:
        actions = copy.deepcopy(actions)

        last_action = None
        new_actions = []

        for i, action in enumerate(actions):
            if action != last_action or type(action) in [Stir, Reflux, Microwave]:
                new_actions.append(action)
            last_action = action

        return new_actions



from typing import Iterator, List, Optional, Sequence

import textdistance
from nltk.translate.bleu_score import corpus_bleu

from .conversion_utils import ActionStringConversionError
from .converter_interface import ActionStringConverter
from .readable_converter import ReadableConverter
from .utils import all_identical


def highlight_differences(
    source_sentences: List[str], translations: Sequence[List[str]]
) -> None:
    """
    Will highlight sentences that are translated differently by different models.

    Args:
        source_sentences: Sentences to translate (length: L)
        translations: Multiple lists of translations, depending on the number of translation models (size: n_models x L)
    """
    assert all(len(t) == len(source_sentences) for t in translations)

    for i, sentence in enumerate(source_sentences):
        sentence_translations = [t[i] for t in translations]

        if not all_identical(sentence_translations):
            print(f"Sample {i}: {sentence}")
            for model_no, s in enumerate(sentence_translations, 1):
                print(f"{model_no}) {s}")
            print()


def full_sentence_accuracy(truth: List[str], pred: List[str]) -> float:
    """
    Calculate the number of exact matches.
    """
    assert len(truth) == len(pred)

    correct_count = sum(int(t == p) for t, p in zip(truth, pred))
    return correct_count / len(truth)


def action_string_validity(
    preds: List[str], converter: Optional[ActionStringConverter] = None
) -> float:
    """
    Calculate the fraction of sentences that correspond to valid action strings.

    Returns:
        value between 0 (no string is valid) and 1.
    """

    if converter is None:
        converter = ReadableConverter()

    n_valid = 0
    for p in preds:
        try:
            converter.string_to_actions(p)
            n_valid += 1
        except ActionStringConversionError:
            pass

    return n_valid / len(preds)


def modified_bleu(truth: List[str], pred: List[str]) -> float:
    """
    Calculates the BLEU score of a translation, with a small modification in order not to penalize sentences
    with less than 4 words.

    Returns:
        value between 0 and 1.
    """
    references = [sentence.split() for sentence in truth]
    candidates = [sentence.split() for sentence in pred]

    # BLEU penalizes sentences with only one word. Even correct translations get a score of zero.
    references = [r + max(0, 4 - len(r)) * [""] for r in references]
    candidates = [c + max(0, 4 - len(c)) * [""] for c in candidates]

    # references must have a larger depth because it supports multiple choices
    refs = [[r] for r in references]
    return corpus_bleu(refs, candidates)  # type: ignore[no-any-return]


def original_bleu(truth: List[str], pred: List[str]) -> float:
    """
    Calculates the BLEU score of a translation, with the original function from nltk.

    Returns:
        value between 0 and 1.
    """
    references = [sentence.split() for sentence in truth]
    candidates = [sentence.split() for sentence in pred]

    # references must have a larger depth because it supports multiple choices
    refs = [[r] for r in references]
    return corpus_bleu(refs, candidates)  # type: ignore[no-any-return]


def levenshtein_similarity(truth: List[str], pred: List[str]) -> float:
    assert len(truth) == len(pred)
    scores: Iterator[float] = (
        textdistance.levenshtein.normalized_similarity(t, p)
        for t, p in zip(truth, pred)
    )
    return sum(scores) / len(truth)


def partial_accuracy(truth: List[str], pred: List[str], threshold: float) -> float:
    """
    Calculates the accuracy from the fraction of sentences that have a similarity to the
    ground truth higher than a given threshold.

    For threshold == 1.0, this function is equivalent to full_sentence_accuracy.

    Args:
        truth: ground truth action sequences
        pred: predicted truth action sequences
        threshold: threshold above which to consider it as a partial match, between 0 and 1
    """
    assert len(truth) == len(pred)
    match_count = sum(
        1
        for t, p in zip(truth, pred)
        if textdistance.levenshtein.normalized_similarity(t, p) >= threshold
    )
    return match_count / len(truth)



import click

from paragraph2actions.sentencepiece_tokenizer import SentencePieceTokenizer


@click.command()
@click.option("--model", "-m", required=True, help="SentencePiece model path")
@click.option("--input_filename", "-i", required=True, help="File to (de)tokenize")
@click.option(
    "--output_filename", "-o", required=True, help="Where to save (de)tokenized text"
)
@click.option("--reverse", "-r", is_flag=True, help="If given, will do detokenization")
def main(model: str, input_filename: str, output_filename: str, reverse: bool) -> None:
    """Tokenize / detokenize with sentencepiece"""

    sp = SentencePieceTokenizer(model)

    with open(input_filename, "rt", encoding="utf-8") as f_in, open(
        output_filename, "wt", encoding="utf-8"
    ) as f_out:
        for line in f_in:
            line = line.strip()
            if reverse:
                detokenized = sp.detokenize(line)
                f_out.write(f"{detokenized}\n")
            else:
                tokenized = sp.tokenize(line)
                f_out.write(f"{tokenized}\n")


if __name__ == "__main__":
    main()



from typing import List

from .actions import (  # noqa
    PH,
    Action,
    Add,
    Chemical,
    CollectLayer,
    Concentrate,
    Degas,
    DrySolid,
    DrySolution,
    Extract,
    Filter,
    FollowOtherProcedure,
    InvalidAction,
    MakeSolution,
    Microwave,
    NoAction,
    OtherLanguage,
    Partition,
    PhaseSeparation,
    Purify,
    Quench,
    Recrystallize,
    Reflux,
    SetTemperature,
    Sonicate,
    Stir,
    Triturate,
    Wait,
    Wash,
    Yield,
)
from .conversion_utils import StringToActionError
from .converter_interface import ActionStringConverter
from .utils import get_all_action_types


class ReprConverter(ActionStringConverter):
    """
    Action-String converter based on the `repr` form of the action instances.

    Not really used anywhere.
    """

    def action_type_supported(self, action_type: str) -> bool:
        return action_type in get_all_action_types()

    def actions_to_string(self, actions: List[Action]) -> str:
        return repr(actions)

    def string_to_actions(
        self,
        action_string: str,
        wrap_errors_into_invalidaction: bool = False,
    ) -> List[Action]:
        try:
            return eval(action_string)  # type: ignore[no-any-return]
        except Exception as e:
            if wrap_errors_into_invalidaction:
                return [InvalidAction(str(e))]
            raise StringToActionError(action_string) from e



from typing import Iterable, Iterator, List, Union

from rxn.onmt_utils.internal_translation_utils import TranslationResult
from rxn.onmt_utils.translator import Translator as RawTranslator

from .sentencepiece_tokenizer import SentencePieceTokenizer


class Translator:
    """
    Wraps the OpenNMT translation functionality into a class.
    """

    def __init__(
        self, translation_model: Union[str, Iterable[str]], sentencepiece_model: str
    ):
        """
        Args:
            translation_model: path to the translation model file(s). If multiple are given, will be an ensemble model.
            sentencepiece_model: path to the sentencepiece model file
        """
        self.sp = SentencePieceTokenizer(sentencepiece_model)
        self.onmt_translator = RawTranslator.from_model_path(translation_model)

    def translate_single(self, sentence: str) -> str:
        """
        Translate one single sentence.
        """
        translations = self.translate_sentences([sentence])
        assert len(translations) == 1
        return translations[0]

    def translate_sentences(self, sentences: List[str]) -> List[str]:
        """
        Translate multiple sentences.
        """
        translations = self.translate_multiple_with_scores(sentences)
        return [t[0].text for t in translations]

    def translate_multiple_with_scores(
        self, sentences: List[str], n_best: int = 1
    ) -> Iterator[List[TranslationResult]]:
        tokenized_sentences = [self.sp.tokenize(s) for s in sentences]

        translations = self.onmt_translator.translate_multiple_with_scores(
            tokenized_sentences, n_best
        )

        for translation_group in translations:
            for t in translation_group:
                t.text = self.sp.detokenize(t.text)
            yield translation_group



import copy
import re
from collections import defaultdict
from typing import Dict, List

from paragraph2actions.misc import TextWithActions

from ..actions import Chemical
from ..utils import extract_chemicals
from .substitution_augmenter import SubstitutionAugmenter


class CompoundNameAugmenter(SubstitutionAugmenter):
    """
    Augments data by substituting compound names.
    """

    def __init__(self, probability: float, compounds: List[str]):
        """
        Args:
            probability: probability with which to switch the compound name
            compounds: list of names to use for substitution
        """
        super().__init__(probability=probability, values=compounds)

    def augment(self, sample: TextWithActions) -> TextWithActions:
        sample = copy.deepcopy(sample)

        chemicals = extract_chemicals(sample.actions)

        # Build a dictionary of compound names and associated chemicals
        # (necessary if the same chemical is present twice)
        cpd_dict: Dict[str, List[Chemical]] = defaultdict(list)
        for c in chemicals:
            cpd_dict[c.name].append(c)

        # remove compound names that are comprised in others; with this, if both '3-ethyltoluene' and
        # '2-bromo-3-ethyltoluene' are present as compounds, we will never substitute the short one.
        for chemical_name in list(cpd_dict.keys()):
            if any(
                chemical_name in cpd for cpd in cpd_dict.keys() if chemical_name != cpd
            ):
                cpd_dict.pop(chemical_name)

        # For each chemical name, try substitution
        for cpd_name in cpd_dict:
            if not self.random_draw_passes() or cpd_name not in sample.text:
                continue
            new_name = self.draw_value()
            sample.text = self.replace_in_text(
                text=sample.text, compound=cpd_name, new_name=new_name
            )
            for c in cpd_dict[cpd_name]:
                c.name = new_name

        return sample

    def replace_in_text(self, text: str, compound: str, new_name: str) -> str:
        # We replace only at word boundaries, to avoid things like 'H2SO4 -> waterSO4' when replacing 'H2' by 'water'
        pattern = re.compile(rf"\b{re.escape(compound)}\b")
        return pattern.sub(new_name, text)



import random
from pathlib import Path
from typing import Callable, List, Set

import click

from paragraph2actions.actions import (
    PH,
    Filter,
    FollowOtherProcedure,
    InvalidAction,
    NoAction,
    Stir,
)
from paragraph2actions.misc import TextWithActions, load_samples, save_samples
from paragraph2actions.readable_converter import ReadableConverter
from paragraph2actions.utils import extract_chemicals


def is_empty(sample: TextWithActions) -> bool:
    return len(sample.actions) == 0


def is_noaction(sample: TextWithActions) -> bool:
    return len(sample.actions) == 1 and isinstance(sample.actions[0], NoAction)


def includes_invalid_action(sample: TextWithActions) -> bool:
    return any(isinstance(a, InvalidAction) for a in sample.actions)


def contains_followed_by(sample: TextWithActions) -> bool:
    return "followed by" in sample.text


def contains_allowed_to(sample: TextWithActions) -> bool:
    return "allowed to" in sample.text


def contains_permitted_to(sample: TextWithActions) -> bool:
    return "permitted to" in sample.text


def contains_left_to(sample: TextWithActions) -> bool:
    return "left to" in sample.text


def contains_saturated(sample: TextWithActions) -> bool:
    return "saturated" in sample.text


def contains_under(sample: TextWithActions) -> bool:
    return " under " in sample.text and not any(
        w in sample.text for w in ["under vac", "reduced pressure"]
    )


def contains_maintain(sample: TextWithActions) -> bool:
    return "maintain" in sample.text


def contains_ice(sample: TextWithActions) -> bool:
    return " ice" in sample.text


def contains_bath(sample: TextWithActions) -> bool:
    return "bath" in sample.text


def contains_neutralize(sample: TextWithActions) -> bool:
    return "neutraliz" in sample.text


def starts_with_after(sample: TextWithActions) -> bool:
    return sample.text.startswith("After")


def includes_unclear_filtration(sample: TextWithActions) -> bool:
    for a in sample.actions:
        if isinstance(a, Filter):
            if a.phase_to_keep is None:
                return True
    return False


def includes_duplicated_chemicals(sample: TextWithActions) -> bool:
    chemicals = extract_chemicals(sample.actions)
    reprs = [repr(c) for c in chemicals]
    return len(chemicals) != len(set(reprs))


def includes_ph(sample: TextWithActions) -> bool:
    return " pH " in sample.text or any(isinstance(a, PH) for a in sample.actions)


def lonely_stir(sample: TextWithActions) -> bool:
    stir_actions = [a for a in sample.actions if isinstance(a, Stir)]
    lonely_stirs = [
        a for a in stir_actions if a.duration is None and a.temperature is None
    ]
    return len(lonely_stirs) != 0


def maybe_follow_other_procedure(sample: TextWithActions) -> bool:
    # do not include the ones already marked to be following other procedures
    if any(isinstance(a, FollowOtherProcedure) for a in sample.actions):
        return False

    return any(w in sample.text.lower() for w in ["prepared", "synthesi"])


subsets: List[Callable[[TextWithActions], bool]] = [
    is_empty,
    is_noaction,
    includes_invalid_action,
    contains_followed_by,
    contains_allowed_to,
    contains_permitted_to,
    contains_left_to,
    contains_saturated,
    contains_under,
    contains_maintain,
    contains_ice,
    contains_bath,
    contains_neutralize,
    starts_with_after,
    includes_unclear_filtration,
    includes_ph,
    includes_duplicated_chemicals,
    lonely_stir,
    maybe_follow_other_procedure,
]


def select_samples(
    samples: List[TextWithActions], n_per_subset: int = 500, n_from_random: int = 2000
) -> List[TextWithActions]:
    # for each subset, get the corresponding samples
    samples_for_subsets = [
        [sample for sample in samples if predicate(sample)] for predicate in subsets
    ]
    print("Total filtered samples:", sum(len(s) for s in samples_for_subsets))

    # Take a given number of samples from each subset and put them together
    merged_samples = []
    for sample_list in samples_for_subsets:
        merged_samples.extend(sample_list[:n_per_subset])
    print(
        f"Size after keeping max {n_per_subset} for each category: {len(merged_samples)}"
    )

    # extend with random (normal) samples
    merged_samples.extend(samples[:n_from_random])
    print("Size after addition of random samples:", len(merged_samples))

    # Remove duplicates
    seen_sentences: Set[str] = set()
    unique_samples: List[TextWithActions] = []
    for sample in merged_samples:
        if sample.text not in seen_sentences:
            seen_sentences.add(sample.text)
            unique_samples.append(sample)

    # shuffle
    random.shuffle(unique_samples)

    print("Size after removing identical sentences:", len(unique_samples))
    return unique_samples


@click.command()
@click.option(
    "--src_in",
    required=True,
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    help="File containing original sentences",
)
@click.option(
    "--tgt_in",
    required=True,
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    help="File containing original sequences",
)
@click.option(
    "--src_out",
    required=True,
    type=click.Path(writable=True, path_type=Path),
    help="Where to save sentences selected for annotation",
)
@click.option(
    "--tgt_out",
    required=True,
    type=click.Path(writable=True, path_type=Path),
    help="Where to save sequences selected for annotation",
)
def main(src_in: Path, tgt_in: Path, src_out: Path, tgt_out: Path) -> None:
    """Generate samples for annotation"""
    action_string_converter = ReadableConverter()

    # Load the data
    samples = load_samples(
        text_file=src_in, actions_file=tgt_in, converter=action_string_converter
    )

    selected_samples = select_samples(samples)

    save_samples(
        samples=selected_samples,
        converter=action_string_converter,
        text_file=src_out,
        actions_file=tgt_out,
    )


if __name__ == "__main__":
    main()



import re
from typing import Callable, Generic, Iterable, List, Optional, Tuple, Type, TypeVar

import attr

from .actions import Action, Chemical

ActionType = TypeVar("ActionType", bound=Action)


class ActionStringConversionError(ValueError):
    """Exception raised for errors during conversion of actions to or from strings."""


class StringToActionError(ActionStringConversionError):
    """Exception raised when the error occurs in the string-to-action conversion."""

    def __init__(self, action_string: str, message: Optional[str] = None):
        self.action_string = action_string
        if message is None:
            message = f'Conversion from string to actions failed for "{action_string}".'
        super().__init__(message)


class ActionToStringError(ActionStringConversionError):
    """Exception raised when the error occurs in the action-to-string conversion."""


class UnsupportedType(ActionToStringError):
    def __init__(self, action_type: Type[Action]):
        self.unsupported_type = action_type.__name__
        super().__init__()


def uppercase_action_name(action: Action) -> str:
    return action.action_name.upper()


def uppercase_type_name(action_type: Type[Action]) -> str:
    return action_type.__name__.upper()


def get_property_from_split(sentence: str, prefix: str) -> Tuple[str, Optional[str]]:
    """
    Gets an optional property following a given prefix.

    Returns:
        Tuple: (sentence before splitting word, optional property)
    """
    separator = f"{prefix} "
    splits = sentence.split(separator)
    prop = separator.join(splits[1:]) if len(splits) > 1 else None
    return splits[0], prop


def chemicals_to_text(chemicals: Iterable[Chemical]) -> str:
    """Get a string representation for Chemical instances."""
    return " and ".join(chemical_to_string(c) for c in chemicals)


def text_to_chemicals(chemicals_text: str) -> List[Chemical]:
    """Get Chemical instances from their string representation."""
    chemicals_strings = chemicals_text.split(" and ")
    return [get_chemical(chemical_string) for chemical_string in chemicals_strings]


def get_chemical(text: str) -> Chemical:
    """
    Convert a string to a chemical.

    Input examples:
        4-butyloctane (5 ml)
        DMF
    """
    remaining, quantities = get_quantities(text)

    # replace the insecable space before the parenthesis back to a normal space
    compound_name = remaining.replace(" \u200C(", " (")

    return Chemical(name=compound_name, quantity=quantities)


def get_quantities(sentence: str) -> Tuple[str, List[str]]:
    match = re.match(r"^.*( \(.*\))$", sentence)
    if not match:
        return sentence, []
    full_match_string = match.group(1)
    # remove parentheses
    match_string = full_match_string[2:-1]
    quantities = match_string.split(", ")
    remaining = sentence.replace(full_match_string, "")
    return remaining, quantities


def chemical_to_string(c: Chemical) -> str:
    """
    Convert a chemical to a string
    """
    # if there is a parenthesis after a space, it may mess up for the
    # back-conversion of the quantities
    compound_name = c.name.replace(" (", " \u200C(")

    if not c.quantity:
        return compound_name
    else:
        quantity_str = ", ".join(c.quantity)
        return f"{compound_name} ({quantity_str})"


@attr.s(auto_attribs=True)
class SingleActionConverter(Generic[ActionType]):
    """
    Object to define the action-to-string and string-to-action conversion for
    a given action type.

    It is used in the ReadableConverter and relies on the principle that an action
    string will consist of the uppercase action type, followed by different
    segments for the associated properties.
    """

    action_type: Type[ActionType]
    action_to_text: Callable[[ActionType], str]
    text_to_action: Callable[[str], ActionType]

    def first_word(self) -> str:
        return uppercase_type_name(self.action_type)



from pathlib import Path
from typing import Tuple

import click
from rxn.utilities.files import dump_list_to_file, load_list_from_file

from paragraph2actions.translator import Translator


@click.command()
@click.option(
    "--translation_models",
    "-t",
    multiple=True,
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    help="Translation model file. If multiple are given, will be an ensemble model.",
)
@click.option(
    "--sentencepiece_model",
    "-p",
    required=True,
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    help="SentencePiece model file",
)
@click.option(
    "--src_file",
    "-s",
    required=True,
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    help="File to translate",
)
@click.option(
    "--output_file",
    "-o",
    required=True,
    type=click.Path(writable=True, path_type=Path),
    help="Where to save translation",
)
def main(
    translation_models: Tuple[Path, ...],
    sentencepiece_model: Path,
    src_file: Path,
    output_file: Path,
) -> None:
    """
    Translate a text with an OpenNMT model.

    This script is derived from the one in the OpenNMT-Py repository, and adds pre-processing and post-processing
    in the form of tokenization and de-tokenization with sentencepiece.
    """
    translator = Translator(
        translation_model=[str(m) for m in translation_models],
        sentencepiece_model=str(sentencepiece_model),
    )

    sentences = load_list_from_file(src_file)
    translations = translator.translate_sentences(sentences)
    dump_list_to_file(translations, output_file)


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
from typing import List

from .actions import Action


class ActionStringConverter(ABC):
    """
    Base class for the conversion of our custom action classes to and from strings.
    """

    @abstractmethod
    def action_type_supported(self, action_type: str) -> bool:
        """
        Whether the conversion to and from a given action type is supported by the converter.
        """

    @abstractmethod
    def actions_to_string(self, actions: List[Action]) -> str:
        """
        Converts a list of actions to a string representation.
        """

    @abstractmethod
    def string_to_actions(
        self, action_string: str, wrap_errors_into_invalidaction: bool = False
    ) -> List[Action]:
        """
        Converts a string representation to the corresponding list of actions.

        Args:
            action_string: string to convert to a list of actions
            wrap_errors_into_invalidaction: if True, exceptions raised during the
                conversion will be converted into an InvalidAction, without propagating
                the exception. Useful for instance to show feedback to a user if a
                single action failed out of many (in the UI).

        Raises:
            ActionStringConversionError.
        """



from pathlib import Path
from typing import Tuple

import click
import sentencepiece as spm


@click.command()
@click.option(
    "--inputs",
    "-i",
    multiple=True,
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    help="Input file(s) on which to train sentencepiece",
)
@click.option(
    "--model",
    "-m",
    required=True,
    type=click.Path(writable=True, path_type=Path),
    help="Where to save the sentencepiece model",
)
@click.option("--vocab_size", "-v", default=16000, type=int, help="Vocabulary size")
def main(inputs: Tuple[Path, ...], model: Path, vocab_size: int) -> None:
    """Learn sentencepiece model"""
    input_files = ",".join(str(p) for p in inputs)
    spm.SentencePieceTrainer.Train(
        f"--input={input_files} --model_prefix={model} "
        f"--vocab_size={vocab_size} --character_coverage=1.0 "
        f"--normalization_rule_name=identity"
    )


if __name__ == "__main__":
    main()



import logging
from typing import List

import chemdataextractor
from chemdataextractor.data import Package

from .sentence_splitter import SentenceSplitter, SentenceSplittingError

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class CdeSplitter(SentenceSplitter):
    """
    Splits sentences using the CDE library.
    """

    def __init__(self, split_sentences_at_newlines: bool = True):
        super().__init__(split_sentences_at_newlines=split_sentences_at_newlines)
        download_cde_data()

    def _split_impl(self, text: str) -> List[str]:
        try:
            paragraph = chemdataextractor.doc.Paragraph(text)
            return [sentence.text for sentence in paragraph.sentences]
        except Exception as e:
            raise SentenceSplittingError(text) from e


def download_cde_data() -> None:
    """Explicitly download the CDE model necessary for splitting sentences, if needed."""
    package = Package("models/punkt_chem-1.0.pickle")
    if package.local_exists():
        return

    logger.info("Downloading the necessary ChemDataExtractor data...")
    package.download()
    logger.info("Downloading the necessary ChemDataExtractor data... Done.")



"""Package setup."""

__copyright__ = """
LICENSED INTERNAL CODE. PROPERTY OF IBM.
IBM Research Licensed Internal Code
(C) Copyright IBM Corp. 2023
ALL RIGHTS RESERVED
"""

from setuptools import setup

setup()



import re
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import attr
from rxn.utilities.strings import remove_postfix, remove_prefix

from .actions import (
    PH,
    Add,
    Chemical,
    CollectLayer,
    Concentrate,
    Degas,
    DrySolid,
    DrySolution,
    Extract,
    Filter,
    FollowOtherProcedure,
    InvalidAction,
    MakeSolution,
    Microwave,
    NoAction,
    OtherLanguage,
    Partition,
    PhaseSeparation,
    Purify,
    Quench,
    Recrystallize,
    Reflux,
    SetTemperature,
    Sonicate,
    Stir,
    Triturate,
    Wait,
    Wash,
    Yield,
)
from .conversion_utils import (
    ActionType,
    SingleActionConverter,
    chemical_to_string,
    chemicals_to_text,
    get_chemical,
    get_property_from_split,
    text_to_chemicals,
    uppercase_action_name,
    uppercase_type_name,
)
from .introspection import get_variable_type


def action_without_parameters(
    action_type: Type[ActionType],
) -> SingleActionConverter[ActionType]:
    """Create an action converter for an action type that does not have any
    parameters.

    This will lead to a conversion relying on the uppercase action name.

    Args:
        action_type: action type to create the converter for.
    """

    def action_to_text(action: ActionType) -> str:
        return uppercase_action_name(action)

    def text_to_action(action_string: str) -> ActionType:
        uppercase_name = uppercase_type_name(action_type)
        if action_string != uppercase_name:
            raise ValueError(
                f'Expected "{uppercase_name}", but the action string is "{action_string}".'
            )
        return action_type()

    return SingleActionConverter(action_type, action_to_text, text_to_action)


@attr.s(auto_attribs=True)
class Parameters:
    """
    Utility class for the automatic generation of an action converter with
    parameters.

    Instances of this class specify the name of the action attribute together
    with a separator / prefix. For more complex transformations, functions
    to and from text can be given.

    Attributes:
        attribute: name of the attribute associated with the action class.
        separator: separator word used in the string conversion, such as "at", "with", etc.
        to_text: can be given if the conversion does not simply rely on a
            separator / prefix. Callback to convert the attribute value to
            the segment of string related to that attribute.
        from_text: can be given if the conversion does not simply rely on a
            separator / prefix. Callback to convert the remaining action string
            to what remains after the attribute extraction and the extracted value.
    """

    attribute: str
    separator: Optional[str]
    to_text: Optional[Callable[[Any], str]] = None
    from_text: Optional[Callable[[str], Tuple[str, Any]]] = None

    @property
    def prefix(self) -> str:
        """
        Helper function to give whatever needs to be added before the actual
        value in the string, independently of whether there is a separator or not.
        """
        if self.separator is None:
            return ""
        return f" {self.separator}"


def action_with_parameters(
    action_type: Type[ActionType],
    parameters_raw: List[Union[Parameters, Tuple[Any, ...]]],
    compound_parameter: Optional[Parameters] = None,
) -> SingleActionConverter[ActionType]:
    """Create an action converter for an action type by automating the generation
    of the text linked to its associated properties.

    This will lead to a conversion relying on the uppercase action name, followed
    by a few portions for each of the parameters.
    This is illustrated very well in the uses of this function for the default
    converters.

    Args:
        action_type: action type to create the converter for.
        parameters_raw: List of parameters to specifying how to convert the
            properties associated with an action to and from text. If provided
            as tuples (for ease of use), will be converted directly to
            Parameters instances.
        compound_parameter: Parameter for a Chemical instance. Will always be
            the first one given in the string after the action name. Given as a
            separate parameter because it is usually feasible to parse it only
            if one knows exactly what comes before in the string (i.e.: action
            name).
    """
    parameters = [
        p if isinstance(p, Parameters) else Parameters(*p) for p in parameters_raw
    ]
    uppercase_name = uppercase_type_name(action_type)

    def action_to_text(action: ActionType) -> str:
        s = uppercase_name

        if compound_parameter is not None:
            value = getattr(action, compound_parameter.attribute)
            if value is not None:
                s += f"{compound_parameter.prefix} {chemical_to_string(value)}"

        for parameter in parameters:
            value_type = get_variable_type(action_type, parameter.attribute)
            value = getattr(action, parameter.attribute)
            if parameter.to_text is not None:
                s += parameter.to_text(value)
            elif value_type is str:
                s += f"{parameter.prefix} {value}"
            elif value_type is Optional[str]:
                if value is not None:
                    s += f"{parameter.prefix} {value}"
            elif value_type is bool:
                if value is True:
                    s += f"{parameter.prefix}"
            elif value_type is Chemical or value_type is Optional[Chemical]:
                if value is not None:
                    s += f"{parameter.prefix} {chemical_to_string(value)}"
            else:
                raise ValueError(f"Cannot convert type {value_type.__name__}")
        return s

    def text_to_action(action_text: str) -> ActionType:
        remaining = action_text
        properties: Dict[str, Any] = {}
        for parameter in reversed(parameters):
            value_type = get_variable_type(action_type, parameter.attribute)
            if parameter.from_text is not None:
                remaining, value = parameter.from_text(remaining)
            elif value_type is bool:
                value = remaining.endswith(parameter.prefix)
                remaining = remove_postfix(remaining, parameter.prefix)
            elif value_type is str:
                remaining, value = get_property_from_split(remaining, parameter.prefix)
                if value is None:
                    raise ValueError("Expected a string, but received None instead.")
            elif value_type is Chemical or value_type is Optional[Chemical]:
                # Note: the prefix already has an empty space as a prefix
                splits = remaining.split(f"{parameter.prefix} ")
                if len(splits) == 1:
                    value = None
                elif len(splits) > 2:
                    raise ValueError(
                        f"Found more than one occurence of {parameter.prefix}"
                    )
                else:
                    remaining, compound_text = splits
                    value = get_chemical(compound_text)
            else:
                remaining, value = get_property_from_split(remaining, parameter.prefix)
            properties[parameter.attribute] = value

        if compound_parameter is not None:
            value_type = get_variable_type(action_type, compound_parameter.attribute)
            if value_type is Optional[Chemical] and remaining == uppercase_name:
                pass
            else:
                # Remove action name and prefix to get the compound text
                compound_text = remove_prefix(
                    remaining,
                    f"{uppercase_name}{compound_parameter.prefix} ",
                    raise_if_missing=True,
                )
                value = get_chemical(compound_text)
                properties[compound_parameter.attribute] = value
                # Remaining afterwards: only the uppercase action name
                remaining = uppercase_name

        if remaining != uppercase_name:
            raise ValueError(
                f'Only "{uppercase_name}" should remain. Actual: "{remaining}".'
            )

        return action_type(**properties)

    return SingleActionConverter(action_type, action_to_text, text_to_action)


def repetition_to_text(repetitions: int) -> str:
    """Convert a number of repetitions to its string representation, to be used
    as the to_text variable of the Parameters class."""
    if repetitions == 1:
        return ""
    return f" {repetitions} x"


def repetition_from_text(remaining_text: str) -> Tuple[str, int]:
    """Extract the number of repetitions from an action string, to be used
    as the from_text variable of the Parameters class."""
    match = re.findall(r"( (\d+) x)$", remaining_text)
    if not match:
        return remaining_text, 1

    # NB: there will be only one match, as the regex contains '$'
    all_matched, digit_matched = match[0]
    remaining = remove_postfix(remaining_text, all_matched, raise_if_missing=True)
    return remaining, int(digit_matched)


def makesolution_to_text(action: MakeSolution) -> str:
    """To be used in the creation of the ActionConverter for MakeSolution."""
    return (
        uppercase_action_name(action) + " with " + chemicals_to_text(action.materials)
    )


def makesolution_from_text(action_text: str) -> MakeSolution:
    """To be used in the creation of the ActionConverter for MakeSolution."""
    chemicals_text = action_text.replace("MAKESOLUTION with ", "")
    compounds = text_to_chemicals(chemicals_text)
    return MakeSolution(materials=compounds)


def partition_to_text(action: Partition) -> str:
    """To be used in the creation of the ActionConverter for Partition."""
    return (
        uppercase_action_name(action)
        + " with "
        + chemicals_to_text([action.material_1, action.material_2])
    )


def partition_from_text(action_text: str) -> Partition:
    """To be used in the creation of the ActionConverter for Partition."""
    chemicals_text = action_text.replace("PARTITION with ", "")
    compounds = text_to_chemicals(chemicals_text)
    if len(compounds) != 2:
        raise ValueError(
            f"Found {len(compounds)} compounds associated with Partition action. Must be 2."
        )
    return Partition(material_1=compounds[0], material_2=compounds[1])


add = action_with_parameters(
    Add,
    [
        ("dropwise", "dropwise"),
        ("temperature", "at"),
        ("atmosphere", "under"),
        ("duration", "over"),
    ],
    Parameters("material", None),
)
collect_layer = action_with_parameters(CollectLayer, [("layer", None)])
concentrate = action_without_parameters(Concentrate)
degas = action_with_parameters(
    Degas,
    [
        ("gas", "with"),
        ("duration", "for"),
    ],
)
dry_solid = action_with_parameters(
    DrySolid,
    [
        ("duration", "for"),
        ("temperature", "at"),
        ("atmosphere", "under"),
    ],
)
dry_solution = action_with_parameters(DrySolution, [("material", "over")])
extract = action_with_parameters(
    Extract,
    [("repetitions", None, repetition_to_text, repetition_from_text)],
    Parameters("solvent", "with"),
)
filter_ = action_with_parameters(Filter, [("phase_to_keep", "keep")])
follow_other_procedure = action_without_parameters(FollowOtherProcedure)
invalid_action = action_with_parameters(
    InvalidAction,
    [
        ("error", None),
    ],
)
makesolution = SingleActionConverter(
    MakeSolution, makesolution_to_text, makesolution_from_text
)
microwave = action_with_parameters(
    Microwave,
    [
        ("duration", "for"),
        ("temperature", "at"),
    ],
)
no_action = action_without_parameters(NoAction)
other_language = action_without_parameters(OtherLanguage)
partition = SingleActionConverter(Partition, partition_to_text, partition_from_text)
ph = action_with_parameters(
    PH,
    [("ph", "to pH"), ("dropwise", "dropwise"), ("temperature", "at")],
    Parameters("material", "with"),
)
phase_separation = action_without_parameters(PhaseSeparation)
purify = action_without_parameters(Purify)
quench = action_with_parameters(
    Quench,
    [("dropwise", "dropwise"), ("temperature", "at")],
    Parameters("material", "with"),
)
recrystallize = action_with_parameters(Recrystallize, [], Parameters("solvent", "from"))
reflux = action_with_parameters(
    Reflux,
    [
        ("duration", "for"),
        ("atmosphere", "under"),
        ("dean_stark", "with Dean-Stark apparatus"),
    ],
)
set_temperature = action_with_parameters(SetTemperature, [("temperature", None)])
sonicate = action_with_parameters(
    Sonicate,
    [
        ("duration", "for"),
        ("temperature", "at"),
    ],
)
stir = action_with_parameters(
    Stir,
    [
        ("duration", "for"),
        ("temperature", "at"),
        ("atmosphere", "under"),
    ],
)
triturate = action_with_parameters(Triturate, [], Parameters("solvent", "with"))
wait = action_with_parameters(
    Wait,
    [
        ("duration", "for"),
        ("temperature", "at"),
    ],
)
wash = action_with_parameters(
    Wash,
    [("repetitions", None, repetition_to_text, repetition_from_text)],
    Parameters("material", "with"),
)
yield_ = action_with_parameters(Yield, [], Parameters("material", None))


def default_action_converters() -> List[SingleActionConverter]:  # type: ignore[type-arg]
    """
    Get the default single-action converters for RXN.
    """
    action_converters: List[SingleActionConverter] = [  # type: ignore[type-arg]
        add,
        collect_layer,
        concentrate,
        degas,
        dry_solid,
        dry_solution,
        extract,
        filter_,
        follow_other_procedure,
        invalid_action,
        makesolution,
        microwave,
        no_action,
        other_language,
        partition,
        ph,
        phase_separation,
        purify,
        quench,
        recrystallize,
        reflux,
        set_temperature,
        sonicate,
        stir,
        triturate,
        wait,
        wash,
        yield_,
    ]
    return action_converters



import copy
from typing import List, Set

from paragraph2actions.misc import TextWithActions

from .substitution_augmenter import SubstitutionAugmenter


class ActionAttributeAugmenter(SubstitutionAugmenter):
    """
    Augments data by substituting a given attribute of actions (such as durations or values).
    """

    def __init__(self, probability: float, values: List[str], attribute_name: str):
        """
        Args:
            probability: probability with which to switch the compound name
            values: list of values to use for substitution
            attribute_name: name of the variable of Action classes that contains the values to substitue
        """
        super().__init__(probability=probability, values=values)
        self.attribute_name = attribute_name

    def augment(self, sample: TextWithActions) -> TextWithActions:
        sample = copy.deepcopy(sample)

        values: Set[str] = set()
        for a in sample.actions:
            value = getattr(a, self.attribute_name, None)
            if value is not None:
                values.add(value)

        # remove values that are comprised in others; with this, if both '0 °C' and
        # '10 °C' are present as values, we will never substitute the short one.
        for d in list(values):
            if any(d in value for value in values if d != value):
                values.remove(d)

        # For each quantity, try substitution
        for d in values:
            if not self.random_draw_passes() or d not in sample.text:
                continue

            new_value = self.draw_value()
            sample.text = sample.text.replace(d, new_value)

            for a in sample.actions:
                action_value = getattr(a, self.attribute_name, None)
                if action_value == d:
                    setattr(a, self.attribute_name, new_value)

        return sample



from abc import ABC, abstractmethod
from typing import List


class SentenceSplittingError(ValueError):
    def __init__(self, text_to_split: str):
        super().__init__(
            f"Error when splitting the following text in sentences: {text_to_split}"
        )


class SentenceSplitter(ABC):
    def __init__(
        self,
        split_sentences_at_newlines: bool = True,
        add_full_stop_if_missing: bool = True,
    ):
        """
        Args:
            split_sentences_at_newlines: whether to consider newlines as sentence separators.
            add_full_stop_if_missing: if the paragraph(s) do not have a full stop at the end,
                add it.
        """
        self.split_sentences_at_newlines = split_sentences_at_newlines
        self.add_full_stop_if_missing = add_full_stop_if_missing

    def split(self, text: str) -> List[str]:
        """
        Splits a text into sentences.

        As an example, this is required for the translation model paragraphs->actions:
        the paragraph must be split in sentences before translation.
        """

        # If the text contains newlines: split there already
        if self.split_sentences_at_newlines:
            pre_split_paragraphs = text.splitlines()
        else:
            pre_split_paragraphs = [text]

        sentences = []
        for paragraph in pre_split_paragraphs:
            paragraph = paragraph.strip()
            if not paragraph:
                continue
            if self.add_full_stop_if_missing and not paragraph.endswith("."):
                paragraph += "."
            sentences.extend(self._split_impl(paragraph))

        return sentences

    @abstractmethod
    def _split_impl(self, text: str) -> List[str]:
        """
        Internal, implementation-specific splitting of sentences.

        Applied after any processing in the base class.
        """



import copy
from typing import List, Optional

from ..actions import Action, Add, SetTemperature, Stir, Wait
from .action_postprocessor import ActionPostprocessor


class WaitPostprocessor(ActionPostprocessor):
    def __init__(self) -> None:
        self.ineligible_actions = {Add}

    def postprocess(self, actions: List[Action]) -> List[Action]:
        updated_actions: List[Optional[Action]] = list(copy.deepcopy(actions))

        # go through the actions, "consumed" Wait actions are converted to None in order to remove them afterwards
        for i in range(len(updated_actions) - 1):
            a = updated_actions[i]
            b = updated_actions[i + 1]

            # NB: we don't merge if b has a temperature
            if not isinstance(b, Wait) or b.temperature is not None:
                continue

            # convert SetTemperature to Stir
            if isinstance(a, SetTemperature):
                a = Stir(temperature=a.temperature)

            if not self.eligible_first_action(a):
                continue

            setattr(a, "duration", b.duration)
            updated_actions[i] = a
            updated_actions[i + 1] = None

        return [a for a in updated_actions if a is not None]

    def eligible_first_action(self, a: Optional[Action]) -> bool:
        if any(isinstance(a, cls) for cls in self.ineligible_actions):
            return False

        return hasattr(a, "duration") and getattr(a, "duration") is None



import logging
import random
from pathlib import Path
from typing import Iterable, List

from .actions import InvalidAction
from .conversion_utils import SingleActionConverter
from .misc import TextWithActions, load_samples, load_samples_from_json, save_samples
from .readable_converter import ReadableConverter
from .repr_converter import ReprConverter

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def invalid_filter(sample: TextWithActions) -> TextWithActions:
    """Replace sequences containing (potentially among others) an InvalidAction
    by a sequence with a single InvalidAction without error message."""
    if any(isinstance(a, InvalidAction) for a in sample.actions):
        return TextWithActions(text=sample.text, actions=[InvalidAction()])
    else:
        return sample


def create_annotation_splits(
    annotated_data_dir: Path,
    output_dir: Path,
    single_action_converters: Iterable[SingleActionConverter],  # type: ignore[type-arg]
    valid_fraction: float,
    test_fraction: float,
) -> None:
    repr_converter = ReprConverter()
    readable_converter = ReadableConverter(
        single_action_converters=single_action_converters
    )

    json_file = annotated_data_dir / "annotated.json"
    sents_file = annotated_data_dir / "sentences.txt"
    actns_file = annotated_data_dir / "actions.txt"

    if json_file.exists():
        items = load_samples_from_json(
            annotated_file=json_file,
            converter=readable_converter,
        )
    else:
        items = load_samples(
            text_file=sents_file,
            actions_file=actns_file,
            converter=repr_converter,
        )

    logger.info("Annotation count:", len(items))

    items = [invalid_filter(item) for item in items]

    # remove empty
    items = [item for item in items if len(item.actions) > 0]
    logger.info("Without empty:", len(items))

    # Shuffle
    random.shuffle(items)

    # Training splits
    n_valid_samples = int(valid_fraction * len(items))
    n_test_samples = int(test_fraction * len(items))
    test_items = items[:n_test_samples]
    valid_items = items[n_test_samples : n_test_samples + n_valid_samples]
    train_items = items[n_test_samples + n_valid_samples :]

    logger.info("Samples in train set:", len(train_items))
    logger.info("Samples in valid set:", len(valid_items))
    logger.info("Samples in test set:", len(test_items))

    def save_to_file(sample_list: List[TextWithActions], subset_name: str) -> None:
        src_file = f"{output_dir}/src-{subset_name}.txt"
        tgt_file = f"{output_dir}/tgt-{subset_name}.txt"

        logger.info(f"Saving {len(sample_list)} samples to {src_file} and {tgt_file}")
        save_samples(
            samples=sample_list,
            converter=readable_converter,
            text_file=src_file,
            actions_file=tgt_file,
        )

    output_dir.mkdir(exist_ok=True)
    save_to_file(train_items, "train")
    save_to_file(valid_items, "valid")
    save_to_file(test_items, "test")
